////////////////////////////////////////////////////////////////
// Flash: Intel G18 x16||x16 version
//
//  Flash write- and erase-routines optimized specifically for Intel G18
//  StrataFlash.   And implemented and tested for only one member of
//  the family, at that.
//
//  If you want to know all about this API, see the large, complete
//  comment atop the accomanying header-file.
//
////////////////////////////////////////////////////////////////

#include <io.h>
#include <stdio.h>
#include "flash_intel_g18_x16.h"

////////////////////////////////////////////////////////////////
// It's handy, and not that hard, to keep some statistics on "how long
// did it take me to program that buffer?" and such.
//
// A structure helps organize the statistics we like to keep about the
// programming-process.
////////////////////////////////////////////////////////////////
typedef struct flash_program_stats_struct {
    unsigned int max_buffer_retry_count;
    unsigned int min_buffer_retry_count;
    unsigned int cum_buffer_retry_count;

    unsigned int max_program_retry_count;
    unsigned int min_program_retry_count;
    unsigned int cum_program_retry_count;

    unsigned int num_samples;
} flash_program_stats;


////////////////////////////////////////////////////////////////
// flash_program_stats_init
//
// Initialize the programming-statistics structure.
////////////////////////////////////////////////////////////////
static void flash_program_stats_init(flash_program_stats* fps)
{
    fps->max_buffer_retry_count = 0;
    fps->min_buffer_retry_count = 0xFFFFFFFF;
    fps->cum_buffer_retry_count = 0;

    fps->max_program_retry_count = 0;
    fps->min_program_retry_count = 0xFFFFFFFF;
    fps->cum_program_retry_count = 0;

    fps->num_samples = 0;
}

////////////////////////////////////////////////////////////////
// flash_program_stats_add_sample
//
// Add the buffer-acquisition- and program-done-polling-counts for the
// last operation.
//
// If you call this after every operation, the structure "keeps score"
// and you can ask for your stats later.
//
////////////////////////////////////////////////////////////////
static void flash_program_stats_add_sample (flash_program_stats* fps,
				     unsigned int buffer_retry_count, unsigned int program_retry_count)
{
    fps->num_samples++;
    fps->cum_buffer_retry_count += buffer_retry_count;
    fps->cum_program_retry_count += program_retry_count;

    fps->max_buffer_retry_count = MAX(buffer_retry_count,
            fps->max_buffer_retry_count);
    fps->min_buffer_retry_count = MIN(buffer_retry_count,
            fps->min_buffer_retry_count);
    fps->max_program_retry_count = MAX(program_retry_count,
            fps->max_program_retry_count);
    fps->min_program_retry_count = MIN(program_retry_count,
            fps->min_program_retry_count);
}

////////////////////////////////////////////////////////////////
// flash_program_stats_print
//
// Print-out the score (statistics results), presumably after some
// relatively-large amount of programming activity.
//
////////////////////////////////////////////////////////////////
static void flash_program_stats_print(flash_program_stats* fps)
{
#if FLASH_INTEL_G18_X16_DEBUG
    printf("%d cmnds, %d average retries, range [%d..%d].\n",
            fps->num_samples,
            fps->cum_program_retry_count / fps->num_samples,
            fps->min_program_retry_count,
            fps->max_program_retry_count);
#endif
}

////////////////////////////////////////////////////////////////
// flash_print_block_map
//
// This function prints block map information found by
// flash_init_block_map when FLASH_INTEL_G18_X16_DEBUG is true.
////////////////////////////////////////////////////////////////
static void flash_print_block_map(unsigned int block_size, unsigned int num_blcks)
{
#if FLASH_INTEL_G18_X16_DEBUG
	printf("\n block_size = 0x%x", block_size); // 0x40000
	printf("\n num_blcks  = 0x%x", num_blcks);  // 0x200
#endif
}

////////////////////////////////////////////////////////////////
// flash_init_block_map : block-map [MAX_FLASH_BLOCKS]
//
// This function initializes an array (provided by the caller) with
// the BYTE-offsets of the boundaries of the various blocks.  It
// determines this based on a couple of constants defined in the
// header-file.  See the constants and comments there.
// (SHOULD only need to be made ONCE unless it takes too much mem?)
////////////////////////////////////////////////////////////////
static void flash_init_block_map (int base, unsigned int block_map [MAX_FLASH_BLOCKS])
{
    int boundary_num = 0;
    unsigned int offset = 0;
	unsigned int block_size, num_blcks;

    // num_regions = IORD_8DIRECT( base, 0x58 ); (why would we need region count?)

    IOWR_16DIRECT(base, 0, FLASH_CMD16_QUERY);

	// Calculate the number of blocks from the CFI table data
	// num_blcks = ((IORD_8DIRECT(base, (0x2E*4) ) << 8) | IORD_8DIRECT(base, (0x2D*4))) + 1;
	num_blcks = 0x200;

	// Calculate the block sizes from the CFI table data
	// block_size = (((IORD_8DIRECT(base, (0x30*4)) << 8) | IORD_8DIRECT(base, (0x2F*4))) << 8);
	block_size = 0x40000;

	flash_print_block_map(block_size, num_blcks);

	for (; boundary_num < num_blcks; boundary_num++) {
		block_map [boundary_num] = offset;
		offset += block_size;
#ifdef MAN_DEBUG
		printf("\n block_map[%d] = 0x%x, ", boundary_num, offset); // special debug
#endif
    }
}

////////////////////////////////////////////////////////////////
// flash_find_previous_block_boundary
//
// Given an offset X in the flash, finds the bottom (lowest offset) of
// the block containing X.
//
// This has a (slightly) inefficient implementation, because it
// recreates an entire, private block-map every time you ask the
// question.
//
// In practice, this is no big deal because you're usually only asking
// this question (at most) a few dozen times during a huge erase- or
// program-operation (and usually: more like once).   These operations
// take SECONDS, so the amount of time required to compute a block-map
// a dozen times is negligible.
//
// But you should avoid calling this function in your inner-loop.
//
////////////////////////////////////////////////////////////////
static unsigned int flash_find_previous_block_boundary (unsigned int base, unsigned int offset)
{
    unsigned int block_map [MAX_FLASH_BLOCKS];
    int i;

    flash_init_block_map(base, block_map);

    for (i = 1; i < MAX_FLASH_BLOCKS; i++)
        if (block_map [i] > offset)
            break;

#ifdef MAN_DEBUG
    printf("\n found previous block_map[%d]\n", i); // debug
#endif
    // seems to step twice here then seems to hang
    return block_map [i - 1];
}

////////////////////////////////////////////////////////////////
// flash_find_next_block_boundary
//
// Given an offset X in the flash, finds the bottom (lowest offset) of
// the NEXT address-sequential block.
//
// Like its brother ..._previous_..., this function computes a private
// block-map every time it's called, so all the same caveats apply.
//
////////////////////////////////////////////////////////////////
static unsigned int flash_find_next_block_boundary(unsigned int base, unsigned int offset)
{
    unsigned int block_map [MAX_FLASH_BLOCKS];
    int i;

    flash_init_block_map(base, block_map);

    for (i = 1; i < MAX_FLASH_BLOCKS; i++)
        if (block_map [i] > offset)
            break;

#ifdef MAN_DEBUG
    printf("\n found next block_map[%d]\n", i); // debug
#endif

    return block_map [i];
}


////////////////////////////////////////////////////////////////
// flash_clear_SR_register
//
// According to the datasheet, the G18's WSM (write state-machine) can
// only *set* bits in the status register.  So it's probably a good
// idea to *clear* the status-register before you start an operation.
//
////////////////////////////////////////////////////////////////
static void flash_clear_SR_register (unsigned int base, unsigned int offset)
{
	IOWR_16DIRECT(base, offset, FLASH_CMD16_CLEAR_STATUS);
}

////////////////////////////////////////////////////////////////
// flash_read_mode
//
// Restores the indicated block to "array read" mode.
//
////////////////////////////////////////////////////////////////
static void flash_read_mode (unsigned int base, unsigned int offset)
{
	flash_clear_SR_register(base, offset);
	IOWR_16DIRECT(base, offset, FLASH_CMD16_READ_MODE);
}

////////////////////////////////////////////////////////////////
// flash_print_errors
//
// Detect error bits completely and individually from passed in unsigned char
// Before you can use "fast buffered write" mode, you have to request
// ERS_SUSPEND_8  0x40 // Check SR[bit6]:Erase suspend in effect.                   0100 0000
// CMND_SEQ_8     0x30 // Check SR[b4&5]:Command sequence error - command aborted.  0010 0000
// ERS_ERROR_8    0x20 // Check SR[bit5]:Erase error - operation aborted            0010 0000
// PROG_ERROR_8   0x10 // Check SR[bit4]:Program error - operation aborted.         0001 0000
// VPP_DROPPED_8  0x08 // Check SR[bit3]:VPP = VPPLK during program or erase op.    0000 1000
// PROG_SUSPEND_8 0x04 // Check SR[bit2]:Program suspend in effect.                 0000 0100
// BLK_LOCKED_8   0x02 // Check SR[bit1]:Blk locked during prog or erse; op aborted 0000 0010
// BEFP_INPROG_8  0x01 // Check SR[bit0]:Buffered Enhanced Factory Prog in-progress 0000 0001
//
////////////////////////////////////////////////////////////////
static void flash_print_errors(unsigned short error)
{
    error &= 0x80; // clear bit7

    // first check for bits 4, 5, or both (either or condition)
    if (error & CMND_SEQ_8) // SR[b4&5] 0010 0000
        printf("CMND_SEQ_8:Command sequence error - command aborted. \n");
    else if (error & PROG_ERROR_8) // SR[bit4] 0001 0000
        printf("PROG_ERROR_8:Program error - operation aborted. \n");
    else if (error & ERS_ERROR_8) // SR[bit5] 0010 0000
        printf("ERS_ERROR_8:Erase error - operation aborted. \n");

    // see if there's any other set error bits
    if (error & BLK_LOCKED_8) // SR[bit1] 0000 0010
        printf("BLK_LOCKED_8:Block locked during program or erase; operation aborted. \n");
    if (error & BEFP_INPROG_8) // SR[bit0] 0000 0001
        printf("BEFP_INPROG_8:Buffered Enhanced Factory Programming in-progress. \n");
    if (error & PROG_SUSPEND_8) //SR[bit2] 0000 0100
        printf("PROG_SUSPEND_8 Program suspend in effect.\n");
    if (error & VPP_DROPPED_8) // SR[bit3] 0000 1000
        printf("VPP_DROPPED_8:VPP<VPPLK during program or erase operation. \n");
    if (error & ERS_SUSPEND_8) // SR[bit6] 0100 0000
        printf("ERS_SUSPEND_8:Erase suspend in effect. \n");
    if (error & 0x7fff) // 0111 1111 remaining bits all 1's? // This should never happen!
        printf(" Status = 0xFFFF:Device not in STATUS mode? \n");
}

////////////////////////////////////////////////////////////////
// flash_parse_errors
//
// An error was detected. pass in unsigned int of full 16||16 status
// Split into two distinct 16 bit status shorts.
// Print error status for each devise using flash_print_errors() to
// analyze and print status errors bit-by-bit.
//
////////////////////////////////////////////////////////////////
static void flash_parse_errors(unsigned int sr_val)
{
    volatile unsigned int sr_valL = sr_val & 0x0000FFFF; // put lower byte into sr_valL
    volatile unsigned int sr_valU = sr_val & 0xFFFF0000; // put upper byte into sr_valU
    sr_valU = sr_valU >> 16; // move upper to lower for bit mask checking

    printf("\nOne or more Flash ERRORs detected: ");

    printf("Lower flash device status = 0x%8x\n", sr_valL);
    flash_print_errors((unsigned short) sr_valL);

    printf("Upper flash device status = 0x%8x\n", sr_valU);
    flash_print_errors((unsigned short) sr_valU);
}

////////////////////////////////////////////////////////////////
// flash_wait_for_operation_done
//
// When you issue a command to the flash, you read-back the "status"
// register to see if the operation has finished.  Bit 7 of the
// "status" register indicates "done;" other bits indicate various
// kinds of errors.
//
// This function returns zero if successful (operation finished, error-
// free, within the given timeout-interval), and nonzero if there is
// an error.
//
// A timeout results in a negative return value.
//
// If a device-error occurs, the value of the status-register is
// returned.
//
////////////////////////////////////////////////////////////////
static int flash_wait_for_operation_done(unsigned int base, unsigned int offset, int* retry_count, int retry_max)
{
    *retry_count = 0;
    volatile int done = 0;

    do {
        volatile unsigned int sr_val = IORD_16DIRECT(base, offset);

        // both devices done SR7 (lower) and SR23 (upper) high
        if ((sr_val & FLASH_MASK16_SR_READY) == FLASH_MASK16_SR_READY) // 0x00800080
        {
            done = 1; // set done
            // Done...but were there any errors?
            if (done && (sr_val != FLASH_MASK16_SR_READY)) {
#if FLASH_INTEL_G18_X16_DEBUG
                flash_parse_errors(sr_val); // parse and print errors
                printf("Error waiting for operation:  Error %X after %d retrys\n",
                        sr_val, (int) retry_count); // announce error
#endif
                retry_max = 1000;
                //return sr_val;
            } else {
                // Done, no error.
                return 0;
            }
        }
    } while ((*retry_count)++ < retry_max);

    if (done)
        return 0;
    else
	    return -1;
}

////////////////////////////////////////////////////////////////
// flash_read_device_id_reg
//
// You can access the "device ID" register by sending a
// command-sequence (or by calling this routine which does it for
// you).  This register contains two bits which tell you about the
// lock-status of the G18.
//
// checks to be sure that x16||x16 dev_identifiers are present/matching.
// Emits error if not.
//
////////////////////////////////////////////////////////////////
static int flash_read_config_id(unsigned int base, unsigned int block_offset)
{
    volatile unsigned int device_id_reg = 0;

    IOWR_16DIRECT(base, block_offset + 2, FLASH_CMD16_READ_CONFIG_ID); // CMD, no offset
    device_id_reg = IORD_16DIRECT(base, block_offset + 2); // offsets are in bytes

    if (device_id_reg != FLASH16_DEV_ID) {
        printf("Error: device-id for adjacent x16 devices cannot be read or do not match = 0x%X\n", device_id_reg);
    } else
        device_id_reg = 0; // return status good.

    // if error, return device ID's, return zero
    return device_id_reg;
}

////////////////////////////////////////////////////////////////
// flash_check_lock_status
//
// You can access the "device ID" register by sending a
// command-sequence (or by calling this routine which does it for
// you).  This register contains two bits which tell you about the
// lock-status of the G18.
//
// Block Lock Configuration:
// BBA(1) + 0x02
// Lock Bit:
// � Block Is Unlocked DQ0 			= 0b0
// � Block Is Locked DQ0 			= 0b1
// � Block Is not Locked-Down DQ1 	= 0b0
// � Block Is Locked-Down DQ1 		= 0b1
//
// returns locked_status = clear or set
//
////////////////////////////////////////////////////////////////
static int flash_check_lock_status (unsigned int base, unsigned int block_offset)
{
    volatile unsigned int locked_status = 0;

    // flash_read_config_id() will open the config register
    // and check the device ID's. "Borrows" locked_status variable.
    if (!(locked_status = flash_read_config_id(base, block_offset)))
        locked_status = IORD_16DIRECT(base, (block_offset + 4 * 2)); // offset in bytes
    else {
        printf("Passing dev_id error 0x%X up to calling function...\n", locked_status);
        return locked_status; // pass the dev_id error up
    }

    // make sure both blocks are unlocked (clear), else return flag to clear them
    if (locked_status) {
        //		printf("Alert: one or both x16 flash blocks indicate \"locked\" = 0x%X, Unlocking...\n",
        //				locked_status);
#if FLASH_INTEL_G18_X16_DEBUG
		printf(".LOCK.\n"); // "." non-verbosely indicates unlock condition

        // then parse lock status
        volatile unsigned int locked_statusL = locked_status & 0x0000FFFF; // lower byte-> locked_statusL
        volatile unsigned int locked_statusU = locked_status & 0xFFFF0000; // upper byte-> locked_statusU
        locked_statusU = locked_statusU >> 16; // move upper to lower for bit mask checking

        if (locked_statusL)
            printf("\tdevice-id for lower x16 locked status reads 0x%X\n", locked_statusL);
        if (locked_statusU)
            printf("\tdevice-id for upper x16 locked status reads 0x%X\n", locked_statusU);
#endif
    }

    return locked_status; // should be clear
}

////////////////////////////////////////////////////////////////
// flash_unlock_erase_block
//
// "Unlocks" the erase-block which contains the BYTE-offset "offset."
//
// As-written, this function always assumes the operation was
// successful.  There is some checking we could do if the need ever
// arises.
//
////////////////////////////////////////////////////////////////
static int flash_unlock_erase_block (unsigned int base, unsigned int offset, int retry_max)
{
    unsigned int error_code = 0;
    unsigned int block_start_offset = flash_find_previous_block_boundary(base, offset);

    // check if block is locked,
    error_code = flash_check_lock_status(base, block_start_offset);

    if (error_code) { //skip, unlock, or find error
        volatile int done = 0;
        int* retry_count = 0;

		flash_clear_SR_register(base, offset); // from qry, puts us back in read mode

        // perform unlock
        IOWR_16DIRECT(base, block_start_offset, FLASH_CMD16_BLOCK_LOCK_SETUP); // 0x0060
        IOWR_16DIRECT(base, block_start_offset, FLASH_CMD16_BLOCK_UNLOCK); // 0x00D0

        do {
            volatile unsigned int sr_val = IORD_16DIRECT(base, offset);

            // both devices done SR7 (lower) and SR23 (upper) high
            if ((sr_val & FLASH_MASK16_SR_READY) == FLASH_MASK16_SR_READY) {
                done = 1; // set done

                // Done...but were there any errors?
                if (done && (sr_val != FLASH_MASK16_SR_READY)) {
#if FLASH_INTEL_G18_X16_DEBUG
                    flash_parse_errors(sr_val); // parse and print errors
                    printf("Error waiting after unlock block %X:  Error %X after %d retrys\n",
                            offset, sr_val, (int) retry_count); // announce error
#endif
					flash_read_mode(base, offset);
                    return sr_val;
                } else {
                    // Done, no error.
					flash_read_mode(base, offset);
                    return 0;
                }
            }
        }// close do
        while ((*retry_count)++ < retry_max);

		flash_read_mode(base, offset);
        return -1; // timeout, return error
    } // end if check lock

    // We probably want to do more checking here, if we start to have trouble
    return error_code; // pass up the check lock result
}

////////////////////////////////////////////////////////////////
// flash_acquire_write_buffer
//
// Before you can use "fast buffered write" mode, you have to request
// (and be granted) access to the G18's internal write buffer.  This
// function requests access, waits until it is granted, and tells you
// whether or not it was successful.
//
// "retry_max" is the maximum number of times you want this to poll
// for the buffer to be ready.  "retry_count" will contain the actual
// number of polling-events performed after the function completes.
//
////////////////////////////////////////////////////////////////
static int flash_acquire_write_buffer (alt_u32 base, alt_u32 offset, int* retry_count, int retry_max)
{
    volatile int done = 0;
    *retry_count = 0;

    do {
        IOWR_16DIRECT(base, offset, FLASH_CMD16_WR_BUF);
        volatile unsigned int sr_val = IORD_16DIRECT(base, offset);

        // both devices done SR7 (lower) and SR23 (upper) high
        if ((sr_val & FLASH_MASK16_SR_READY) == FLASH_MASK16_SR_READY) {
            done = 1; // set done

            // Done...but were there any errors?
            if (done && (sr_val != FLASH_MASK16_SR_READY)) {
#if FLASH_INTEL_G18_X16_DEBUG
                flash_parse_errors(sr_val); // parse and print errors
                printf("Error waiting after acquire_write_buffer:  Error %X after %d retrys\n", sr_val, (int) retry_count); // announce error
#endif
                return sr_val;
            } else {
                return 0; // Done, no error.
            }
        }
    } while ((*retry_count)++ < retry_max);

    return -1; // return error
}


////////////////////////////////////////////////////////////////
// flash_fill_write_buffer
//
// Takes "buffer_halfword_count" number of 16-bit words from "buffer" and
// writes them into the G18's write-buffer, in preparation for
// programming those words at "offset."
//
// This function is just a small chunk of "flash_program_buffer."
//
// You must acquire the buffer before you call this function.
//
////////////////////////////////////////////////////////////////
static void flash_fill_write_buffer(unsigned int base, unsigned int offset, unsigned short* buffer, int loop_word_count)
{
	int cnt = loop_word_count * 2;
    while (cnt-- > 0) {
        IOWR_16DIRECT(base, offset, *buffer++);
        offset += 2;
    }
}


////////////////////////////////////////////////////////////////
// flash_program_buffer
//
// Executes ONE complete "fast buffered write" sequence to the flash.
// The result is either:
//
//     * 1..32 words written to the flash successfully, or
//     * an error-code.
//
// It is the caller's responsibility to make sure the region of flash
// at "offset" is erased prior to calling this function.  You may not
// use this function to write more than 32 bytes at a time.
//
// According to the G18 datasheet, this function will run fastest if
// the offset is an even multiple of 32 WORDS (= 64 BYTES).  It is the
// caller's responsibility to keep track of the offset.
//
// You may NOT call this function to program a region which spans a
// block-boundary.  Well: You CAN, but it will result in an error.
//
// Returns zero if successful; negative for timeout; status-register
// value if device error.
//
////////////////////////////////////////////////////////////////

static int flash_program_buffer(unsigned int base, unsigned int offset, unsigned int* buffer, int loop_word_count,
        int* buffer_retry_count, int buffer_retry_max, int* program_retry_count, int program_retry_max)
{
    unsigned int loop_word_count_minus1 = (2*loop_word_count - 1);
   // unsigned int buffer_x16x16_count_minus1 = ((loop_word_count_minus1 & 0xFFFF) | ((loop_word_count_minus1 << 16) & 0xFFFF0000));
#ifdef MAN_DEBUG
    printf("loop_word_count_minus1 = 0x%8x\n", loop_word_count_minus1); // debug 0x     1f
    printf("buffer_x16x16_count_minus1 = 0x%8x\n", buffer_x16x16_count_minus1); // debug 0x  001f 001f
#endif

	flash_clear_SR_register (base, offset);// huh?
	IOWR_16DIRECT(base, offset, FLASH_CMD16_READ_STATUS);
    // Get the write-buffer.
    int error_code = flash_acquire_write_buffer(base, offset, buffer_retry_count, buffer_retry_max);
    if (error_code != 0) {
#if FLASH_INTEL_G18_X16_DEBUG
        printf("Unable to acquire write-buffer (code %X) after %d retries\n", error_code, *buffer_retry_count);
#endif
        return error_code;
    }

    // Write the word-count (minus one, because that's what the intel
    // app-note says clearly; and the spec says very-obtusely:
    IOWR_16DIRECT(base, offset, loop_word_count_minus1); // buffer_halfword_count of each device! 0x001f001f
#if 0
    error_code = IORD_16DIRECT(base, offset);
    if (error_code & FLASH_MASK1616_SR_ERROR) {
#if FLASH_INTEL_G18_DEBUG
        printf("Error (%X) writing word-count (%d) to flash.\n", error_code, buffer_word_count - 1);
#endif
        return error_code;
    }
#endif
    //Fill the flash write-buffer with the words:
    flash_fill_write_buffer(base, offset, buffer, loop_word_count);

    // Write the confirm-command,
    IOWR_16DIRECT(base, offset, FLASH_CMD16_CONFIRM); //

    error_code = flash_wait_for_operation_done(base, offset, program_retry_count, program_retry_max);
    if (error_code != 0) {
#if FLASH_INTEL_G18_X16_DEBUG
        printf("Error (%X) waiting for programming to finish after %d retries.\n", error_code, *program_retry_count);
#endif
        return error_code;
    }
    return 0;
}

////////////////////////////////////////////////////////////////
// flash_program_within_erase_block
//
// Programs a region of flash.  The region of flash:
//
//     * Must have been previously-erased.
//     * Must all lie within the same erase-block.
//
// This breaks the region into multiple  32-word "buffer-sized" chunks
// and calls flash_program_buffer as many times as it takes.  Care is
// taken to align the buffers on 64-BYTE boundaries for faster
// operation.
//
// Keeps running statistics on the number of polling-loops required to
// run each command.
//
// Returns zero if successful, negative if timeout, status-register
// value if device-error.
//
////////////////////////////////////////////////////////////////
static int flash_program_within_erase_block(unsigned int base, unsigned int offset, unsigned int* buffer, int word_count, // byte_count/4 = number of times the write_loop runs; 4 bytes per loop
                                            flash_program_stats* fps, int buffer_retry_max, int program_retry_max)
 {
    int buffer_retry_count = 0;
    int program_retry_count = 0;
    int chunk_size = 0;
    int error_code = 0;

    while (word_count > 0) {
        // Start-off assuming we're going to write a maximum-sized chunk,
        chunk_size = FLASH16_MAX_BUFF_SIZE;

        // Deal with the "odd offset" case.
        if (offset % FLASH16_MAX_BYTE_SIZE != 0) { // offset is in bytes
            // Make this chunk only big-enough to get us to the next multiple of FLASH16_MAX_BYTE_SIZE.
            int next_even_offset = ((offset / FLASH16_MAX_BYTE_SIZE) + 1) * FLASH16_MAX_BYTE_SIZE;
            int bytes_to_write = next_even_offset - offset; // offset if a byte-y thing
#ifdef MAN_DEBUG
            printf("ODD OFFSET CASE: offset= 0x%x, word_count= 0x%x, next_even_offset= 0x%x bytes_to_write= 0x%x\n",
                    offset, word_count, next_even_offset, bytes_to_write);
#endif
            // We write 32-bit ganged 16 bit half-words to the flash...not bytes.
            chunk_size = bytes_to_write / BYTES_PER_WORD;
        }

        // Deal with the "we're at the end" case.
        if (chunk_size > word_count) // if maximum is greater than what's left
        {
            chunk_size = word_count; // just write however many are needed for what's left
#ifdef MAN_DEBUG
            printf("END_CASE: offset= 0x%x, word_count= 0x%x, 0x%x\n", offset, word_count);
#endif
        }

        error_code = flash_program_buffer(base, offset, buffer, chunk_size, &buffer_retry_count,
                     buffer_retry_max, &program_retry_count, program_retry_max);

        if (error_code != 0) {
#if FLASH_INTEL_G18_X16_DEBUG
            printf("Flash programming failed with error-code %X.\n", error_code);
#endif
            return error_code;
        }

        flash_program_stats_add_sample(fps, buffer_retry_count, program_retry_count);

        buffer += chunk_size; // Buffer contains 16-bit WORDS.
        offset += chunk_size* BYTES_PER_WORD; // Offset given in 8-bit BYTES.
        word_count -= chunk_size;
    }

	flash_read_mode (base, offset); //why here? don't need during step-debug
    return 0;
}

////////////////////////////////////////////////////////////////
// flash_program_across_blocks
//
//// Programs "word_count" 16-bit words from "buffer" into the
// Programs "word_count" 32-bit words from "buffer" into the
// contiguous region of flash starting at BYTE-offset "offset."
//
// The region of flash must have been previously erased.
//
// The region of flash can span any number of erase-block boundaries.
//
// This function just calls "flash_program_within_erase_block" for
// every erase-block in the region.
//
////////////////////////////////////////////////////////////////
static int flash_program_across_blocks(unsigned int base, unsigned int offset, unsigned int* buffer, // buffer is  32 bits per fullword
        int word_count, flash_program_stats* fps, int buffer_retry_max, int program_retry_max)
{
    int error_code = 0;


    while (word_count > 0) {
        // Figure-out if we're going to run past the next erase-block-boundary
		int chunk_size = FLASH16_MAX_BLOCK_SIZE / BYTES_PER_WORD; // number of times we will write to flash (= 160 byte_count/4 = 40)
		//printf("ODD OFFSET CASE: offset= 0x%x, word_count= 0x%x\n",offset, word_count);


		if (offset % FLASH16_MAX_BLOCK_SIZE != 0) { // offset is in bytes
			// Make this chunk only big-enough to get us to the next multiple of FLASH16_MAX_BYTE_SIZE.
			int next_even_offset = ((offset / FLASH16_MAX_BLOCK_SIZE) + 1 ) * FLASH16_MAX_BLOCK_SIZE;
			int bytes_to_write	 = next_even_offset - offset; // offset if a byte-y thing
#ifdef MAN_DEBUG
			printf("ODD OFFSET CASE: offset= 0x%x, word_count= 0x%x, next_even_offset= 0x%x bytes_to_write= 0x%x\n",
				  offset, word_count, next_even_offset, bytes_to_write);
#endif

			// We write 32-bit ganged 16 bit half-words to the flash...not bytes.
			chunk_size = bytes_to_write / BYTES_PER_WORD;
		}

		// Deal with the "we're at the end" case.
		if (chunk_size > word_count)  // if maximum is greater than what's left
		{
			chunk_size = word_count; // just write however many are needed for what's left
#ifdef MAN_DEBUG
			printf("END_CASE: offset= 0x%x, word_count= 0x%x, 0x%x\n", offset, word_count);
#endif
		}		

        // Unlock this erase-block: // not entirely sure we need this here?
        error_code = flash_unlock_erase_block(base, offset, program_retry_max);
        if (error_code != 0) {
#if FLASH_INTEL_G18_X16_DEBUG
            printf("Error %X unlocking flash-block at offset %uX.\n", error_code, offset);
#endif
            return error_code;
        }

        error_code = flash_program_within_erase_block(base, offset, buffer,
						chunk_size, fps, buffer_retry_max, program_retry_max);

        if (error_code != 0) {
#if FLASH_INTEL_G18_X16_DEBUG
            printf("Error %X writing flash-block at offset %X.\n", error_code, offset);
#endif
            return error_code;
        }

        buffer += chunk_size; // Buffer contains 32-bit WORDS
        offset += chunk_size* BYTES_PER_WORD; // Offset given in 8-bit BYTES
        word_count -= chunk_size; // 0x20a0000 --> 0x20a00a0 (correct)
    }
    return 0;
}

////////////////////////////////////////////////////////////////
// flash_intel_g18_x16x16_write
//
// Public API routine.
//
//   See large comment in the .h-file.
//
// This function just calls "flash_program_across_blocks," but hides
// the whole business about statistics-keeping, and
// automatically-selects default "timeouts" based on constants defined
// in the header-file.
////////////////////////////////////////////////////////////////
int flash_intel_g18_x16_write(alt_u32 base, alt_u32 offset, alt_u32* buffer, int word_count)
{
    flash_program_stats fps;
    flash_program_stats_init(&fps);

	flash_read_mode(base, offset); // in case some error was leftover

    int error_code = flash_program_across_blocks(base, offset, buffer, word_count, &fps,
            FLASH_INTEL_G18_BUFFER_READY_POLLING_LIMIT, FLASH_INTEL_G18_PROGRAM_DONE_POLLING_LIMIT);

    if (error_code == 0) {
#if FLASH_INTEL_G18_X16_DEBUG
        printf("Programming completed: ");
#endif
        flash_program_stats_print(&fps);
    } else {
#if FLASH_INTEL_G18_X16_DEBUG
        printf("Programming failed with error-code %X.\n", error_code);
#endif
    }

    flash_read_mode(base, offset); // in case some error was leftover
    return error_code;
}


////////////////////////////////////////////////////////////////
// flash_erase_block
//
// Erase the single block of flash which contains the BYTE-offset
// "offset."
//
// Returns zero if successful, negative for timeout, or
// status-register value for device-error.
//
////////////////////////////////////////////////////////////////
static int flash_erase_block(alt_u32 base, alt_u32 offset, int* retry_count, int retry_max)
{
    int error_code;

	flash_clear_SR_register (base, offset); // in case of previous error?

    //printf("Erasing Block: 0x%x\n", base+offset ); // debug
	IOWR_16DIRECT(base, offset, FLASH_CMD16_BLK_ERS_SETUP);
	IOWR_16DIRECT(base, offset, FLASH_CMD16_BLK_ERS_CONFIRM);

    error_code = flash_wait_for_operation_done(base, offset, retry_count, retry_max);
    if (error_code != 0) {
#if FLASH_INTEL_G18_X16_DEBUG
        printf("Error code %X waiting for erase after %d retries.\n",
                error_code, *retry_count);
#endif
        return error_code;
    }

	flash_read_mode (base, offset); //why here? don't need during step-debug
    return 0;
}

////////////////////////////////////////////////////////////////
// flash_erase_across_blocks
//
// Erase all of the blocks which intersect with the given range
// (offset + span).  This just calls "flash_erase_block" enough times
// to do the job.
//
////////////////////////////////////////////////////////////////
static int flash_erase_across_blocks(unsigned int base, unsigned int offset,
        alt_u32 byte_count, flash_program_stats* fps, int erase_retry_max)
{
    int retry_count;
    int error_code;
    alt_u32 end = offset + byte_count;
    int remaining = byte_count;

    while (remaining > 0) {
        error_code = flash_unlock_erase_block(base, offset, erase_retry_max);
        if (error_code != 0) {
#if FLASH_INTEL_G18_X16_DEBUG
            printf("Error %X unlocking flash-block at offset %uX.\n", error_code, offset);
#endif
            return error_code;
        }

        error_code = flash_erase_block(base, offset, &retry_count, erase_retry_max);
        if (error_code != 0) {
#if FLASH_INTEL_G18_X16_DEBUG
            printf("Error %X erasing block at %X.\n", error_code, offset);
#endif
            return error_code;
        }
        flash_program_stats_add_sample(fps, 0, retry_count);

        offset = flash_find_next_block_boundary(base, offset + 1); // indexes the cfi table!
        remaining = end - offset;
    }
    return 0;
}

////////////////////////////////////////////////////////////////
// flash_intel_g18_x16_erase
//
// Public API routine.
//
// See large comment in the .h-file.
//
// This function just calls "flash_erase_across_blocks," but hides
// the whole business about statistics-keeping, and
// automatically-selects default "timeouts" based on constants defined
// in the header-file.
//
////////////////////////////////////////////////////////////////
int flash_intel_g18_x16_erase(alt_u32 base, alt_u32 offset, alt_u32 byte_count)
{
    flash_program_stats fps;
    flash_program_stats_init(&fps);

	flash_read_mode(base, offset); // start with clear and read-mode
    int error_code = 0;

    error_code = flash_erase_across_blocks(base, offset, byte_count, &fps, FLASH_INTEL_G18_ERASE_DONE_POLLING_LIMIT );

    if (error_code == 0) {
#if FLASH_INTEL_G18_X16_DEBUG
        printf("Erase completed: ");
#endif
        flash_program_stats_print(&fps);
    } else {
#if FLASH_INTEL_G18_X16_DEBUG
        printf("Erase failed with error-code %X.\n", error_code);
#endif
    }

	flash_read_mode(base, offset); // Following erase and status check

	return error_code;
}

