#ifndef __FLASH_INTEL_G18_X16_H_DEFINED__
#define __FLASH_INTEL_G18_X16_H_DEFINED__

////////////////////////////////////////////////////////////////
// Flash: Intel G18 x16 version
//
//  Flash write- and erase-routines optimized specifically for Intel G18
//  StrataFlash.   And implemented and tested for only one member of
//  the family, at that.
//
//  If you just want generic, safe flash-management routines:  Boy,
//  are you looking in the wrong place.  You probably want to use the CFI
//  flash routines built-right-in to HAL.  The HAL routines are safe,
//  tested, and flexible.  They'll work with virtually-any CFI flash,
//  no hassle.
//
//  These routines are designed to work ONLY with the Intel-brand
//  G18-family StrataFlash devices.  These have been implemented and
//  tested specifically for only one member of this family:
//
//  Numonyx� Axcell G18-65nm 16 bit data Flash Memory - PC28F512G18BF
//
//       * 128Mbit (16MByte) capacity
//       * 16-bit-wide data
//       * "Bottom Boot" configuration
//
// It's relatively-straightforward for *you* to change some constants
// to make this work with other members of the G18 family...but that
// (and any other use of this code) is at your own risk and on your
// own recognizance.
//
// So...why would you want to use this code?
//
// Because it's FAST, that's why.  The write-routines here use the
// G18's "fast buffered write" mode, which is 20x faster (!) than the
// safe "one-word-at-a-time" mode which the HAL routines wisely use
// (because the HAL routines will work with *any* CFI flash, not just
// the G18).
//
// So: If you've got an Intel XXXXX G18 flash, and you want to write
// data to it in a hurry:  You've come to the right place.
//
// Significantly:  Erasing isn't any faster than normal.  Only
// writing.
//
//
// ----THE API
//
// There are only two functions in the public API:
//
//     flash_intel_g18_erase()         // Erase a region of flash.
//                                     //   May span any number of erase-blocks
//
//     flash_intel_g18_write()         // Write a buffer-full of data
//                                     //   to flash.  May span any
//                                     //   number of erase-blocks.
//
// That's all.  You don't have to "open" your flash-device, or
// initialize it, or anything else.  the only things you can do are
// (a) erase, and (b) write.
//
// You are responsible for erasing a region of flash before you try
// writing to it.  If you try to write to unerased flash, the
// write-routine will fail with an error code.
//
// Both routines return 0 to mean "success," and nonzero for failure.
//
// -----TWEAKS
//
// DEBUGGING
//
// If you compile this module with the preprocessor-symbol:
//
//     FLASH_INTEL_G18_DEBUG
//
// defined, then it will print-out interesting statistics when things
// go right, and semi-informative failure messages when things go
// wrong.
//
//
// OTHER G18 FAMILY MEMBERS
//
// A handful of contstants (right here!) define the
// device-characteristics.
//
// If you want to make this code work with another member of the G18
// family, you need to:
//
//      1) Change the NUM_128K_BLOCKS parameter to agree with the
//         number of full, 128KByte blocks present in your intel G18
//         device.
//
//      2) Choose the BOTTOM_BOOT_BLOCK parameter to agree with your
//         device.  "1" means that the smaller, 32-KByte blocks appear
//         at numerically-low offsets in flash address-space ("the
//         bottom"),  and "0" means the smaller, 32-KByte blocks
//         appear at numerically-high offsets ("the top").
//
//      3) (Maybe) change NUM_32K_BLOCKS to agree with the number of
//         small, 32KByte blocks in your device.  But I think every
//         member of the G18 family has four of these blocks, so
//         you'll probably never change this parameter.
//
//      4) Test it out, and good luck with that.  No guarantees from me.
//
//
//
// TIMING and TIMEOUTS
//
// This code uses a very loosey-goosey timeout scheme.  The program
// just polls the ready-status of the G18 after every erase- or write-
// operation, and gives-up after a certain number of tries (returning
// with an error, of course).
//
// The G18 will complete its operation within some range of wall-clock
// time, according (more-or-less) to the datasheet.  But the number of
// samples we can take in that time depends on lots of stuff:  How
// fast is your Nios running?  How is it connected to the flash?  Is
// anyone else using the flash?  Did you compile your code with
// "-O3?", etc., etc.
//
// Happily: The timeouts don't really matter that much.  But they have
// to be here to keep your program from just going-out-to-lunch and
// never coming back if, for example, the flash isn't even connected.
//
// To balance these requirements (Value doesn't matter; Just prevent
// lock-up), very, very long timeout values are used by default.
// These werre "long" compared to the reasonable amount of time
// required by a 100MHz Nios to program flash.  You can change them if
// you want...but you'll probably never have to.  I just don't want
// you to think that their values were carefully-chosen, or that
// anything-much depends on them.
//
////////////////////////////////////////////////////////////////

#include <alt_types.h>

//#define FLASH_INTEL_G18_X16_DEBUG 1

#if FLASH_INTEL_G18_X16_DEBUG
#include <stdio.h>
#endif

// device ID for Numonyx� Axcell G18-65nm 16bit data Flash Memory - PC28F00AG18xx
#define FLASH16_DEV_ID 0x88B0         // x16 side-by-side

// ganged x16x16 data mode
#define FLASH_CMD16_QUERY             0x0098
#define FLASH_CMD16_WR_BUF            0x00E9
#define FLASH_CMD16_CONFIRM           0x00D0
#define FLASH_CMD16_BLK_ERS_SETUP     0x0020
#define FLASH_CMD16_BLK_ERS_CONFIRM   0x00D0
#define FLASH_CMD16_BLOCK_LOCK_SETUP  0x0060
#define FLASH_CMD16_BLOCK_LOCK        0x0001
#define FLASH_CMD16_BLOCK_UNLOCK      0x00D0
#define FLASH_CMD16_CLEAR_STATUS      0x0050
#define FLASH_CMD16_READ_STATUS       0x0070
#define FLASH_CMD16_READ_MODE         0x00FF
#define FLASH_MASK16_SR_READY         0x0080
#define FLASH_MASK16_SR_ERROR         0x003B
#define FLASH_MASK16_DONE             0x0080
#define FLASH_CMD16_READ_CONFIG_ID   	0x0090 // Read Device Id mode. Subsequent reads output man/device codes,
                                               // Config Reg data, Block Lock status,or OTP Reg data on each DQ[15:0]

// x8 error bit masks
// Only the Clear Status Register command clears SR[1, 3, 4, 5].
// If an error is detected, clear the Status register before
// attempting an erase retry or other error recovery.
//#define VPP_RANGE_8  0x0088  // Check SR[bit3]: 1 = VPP Range Error 1000 1000
//#define CMND_SEQ_8   0x00B0     // Check SR[bit4,bit5]: Both 1 = Command Sequence Error 1011 0000
//#define BLK_ERASE_8  0x00A2     // Check SR[bit5]: 1 = Block Erase Error  1010 0000
//#define BLK_ERASE_8  0x00A0     // Check SR[bit5]: 1 = Block Erase Error  1010 0000
//#define BLK_ERASE_8  0x0092     // Check SR[bit5]: 1 = Block Erase Error  1010 0000
//#define BLK_LOCKED_8 0x0082     // Check SR[bit1]: 1 = Attempted erase of locked block; erase aborted  1000 0010

#define ERS_SUSPEND_8  0x40 // Check SR[bit6]: 1 = Erase suspend in effect.                                0100 0000
#define CMND_SEQ_8     0x30 // Check SR[b4&5]: 1 = Command sequence error - command aborted.               0010 0000
#define ERS_ERROR_8    0x20 // Check SR[bit5]: 1 = Erase error - operation aborted                         0010 0000
#define PROG_ERROR_8   0x10 // Check SR[bit4]: 1 = Program error - operation aborted.                      0001 0000
#define VPP_DROPPED_8  0x08 // Check SR[bit3]: 1 = VPP = VPPLK during program or erase operation.          0000 1000
#define PROG_SUSPEND_8 0x04 // Check SR[bit2]: 1 = Program suspend in effect.                              0000 0100
#define BLK_LOCKED_8   0x02 // Check SR[bit1]: 1 = Block locked during program or erase; operation aborted 0000 0010
#define BEFP_INPROG_8  0x01 // Check SR[bit0]: 1 = Buffered Enhanced Factory Programming in-progress       0000 0001


#define MAX(X,Y) (((X)>(Y))?(X):(Y))
#define MIN(X,Y) (((X)<(Y))?(X):(Y))


#define MAX_FLASH_BLOCKS 512

// Arbitrary "timeout" values.
#define FLASH_INTEL_G18_BUFFER_READY_POLLING_LIMIT 1000000
#define FLASH_INTEL_G18_PROGRAM_DONE_POLLING_LIMIT 1000000
#define FLASH_INTEL_G18_ERASE_DONE_POLLING_LIMIT   100000000

////////////////////////////////////////////////////////////////
// Scalable variables used by buffered flash write routines
//
// Intel calls this "buffered_program - word_count"
//
//
///////////////////////////////////////////////////////////////
#define FLASH16_MAX_BUFF_SIZE 256  //  32( 128) 10:30,
#define FLASH16_MAX_BYTE_SIZE 1024 // 512(2048) 09:55,
#define FLASH16_MAX_BLOCK_SIZE 0x40000  //  32( 128) 10:30,

#define BYTES_PER_WORD 4


////////////////////////////////////////////////////////////////
// flash_intel_g18_write
//
//  Write data, contained in a buffer, to a contiguous region of intel
//  G18 flash.
//
//  The parameters are:
//
//  * base          The (byte-)base-address of the flash memory, just
//                  like you'd find defined as a symbol in your
//                  "system.h" file.
//
//  * offset        The BYTE offset (BYTE! offset) of the first
//                  location in the G18 flash you want programmed.
//                  The offset is relative to the base-address of the flash.
//                  Zero-offset means: the first location in the
//                  flash-device will be programmed, regardless of
//                  where the flash is located in the system memory
//                  space.
//
//                  The offset must be an even number (you cannot
//                  write flash starting at an odd byte address).
//
//
//  * buffer        A pointer to the first word in a buffer of 16-bit
//                  "words" (a "word" for this flash-device is 16
//                  bits).   There must be at least word_count valid
//                  16-bit values in this buffer--because: whatever's
//                  there in memory...it's all gonna get written to
//                  flash.
//
//  * word_count    The number of 16-BIT WORDS that you want written
//                  to flash.  Notice that the starting-offset is
//                  given as a BYTE-address offset, but the size of
//                  the payload-buffer is given in 16-BIT WORDS.  This
//                  may be inconsistent and confusing, but you can't
//                  say I haven't been clear about it.
//
//  The destination-region of flash may span any number of
//  erase-blocks.
//
//  The destination-region of flash must already be erased when you
//  call this function...otherwise you will almost certainly get an
//  error.  This function does not attempt to erase the flash before
//  programming.
//
//  This routine DOES unlock every block it tries to write.
//
////////////////////////////////////////////////////////////////
int flash_intel_g18_x16_write(alt_u32 base, alt_u32 offset, alt_u32* buffer, int word_count);

////////////////////////////////////////////////////////////////
// flash_intel_g18_erase
//
//  Erase all the blocks in an Intel G18 flash which include the
//  specified region.
//
//  Note that if the specified region doesn't start or end on a
//  block-boundary, then a larger, enclosing region of flash will
//  actually be erased (rounded to the next-encompasing block-boundary
//  in either direction).
//
//  The parameters are:
//
//  * base          The (byte-)base-address of the flash memory, just
//                  like you'd find defined as a symbol in your
//                  "system.h" file.
//
//  * offset        The BYTE offset (BYTE! offset) of the lowest
//                  address in the specified region. The entire block
//                  containing this location will be erased.
//                  The offset is relative to the base-address of the flash.
//                  Zero-offset means: the first location in the
//                  flash-device will be erased, regardless of
//                  where the flash is located in the system memory
//                  space.
//
//  * span          The BYTE offset (BYTE! offset) which describes the
//                  total BYTE-address span of the region to be
//                  erased.  The block containing the location
//                  "offset+span" will be erased--along with all preceding
//                  blocks all the way down to "offset."
//
//  This routine DOES unlock every block it tries to erase.
//
////////////////////////////////////////////////////////////////
int flash_intel_g18_x16_erase(alt_u32 base, alt_u32 offset, alt_u32 byte_count);


#endif //__FLASH_INTEL_G18_X16_H_DEFINED__
