// =============================================================================
// File Name      : i2c_read.v
// Module         : i2c_read
// Function       : i2c_read 
// Type           : RTL
// -----------------------------------------------------------------------------
// Update History :
// -----------------------------------------------------------------------------
// =============================================================================
// End Revision
// =============================================================================

// synthesis translate_off
`timescale 1ps / 1ps
// synthesis translate_on

module i2c_read #
(
    parameter    SLAVE_ADDR = 7'h55
)(
    input         sys_clk,             //50MHz
    input         sys_rst_n,
     
    input         i2c_sda_in,
	output        i2c_sda_out,
    output        i2c_scl,
    
    input  [15:0] reg_addr,
    input         cmd_en,
	output [31:0] read_data,
    output        read_valid       
 );

    parameter      IDLE         = 5'b00000,
                   I2C_START    = 5'b00001,
                   I2C_WRITE    = 5'b00010,
				   I2C_RESTART  = 5'b00100,
				   I2C_READ     = 5'b01000,
                   I2C_END      = 5'b10000;

    reg [ 4:0]  next_gen_state;
    reg [ 4:0]  gen_state;
    
    reg [ 8:0]  r_divide_clk;
    reg         r1_divide_clk;
    wire        slow_clk;
    reg [ 1:0]  r_cnt;
    reg [ 4:0]  r_write_data_cnt;
	reg [ 5:0]  r_read_data_cnt;
    reg         r_sda;
    reg         r_scl;
    reg [15:0]  r_reg_addr;
    reg         r_read_valid;
	reg [ 7:0]  r1_read_data;
	reg [ 7:0]  r2_read_data;
	reg [ 7:0]  r3_read_data;
	reg [ 7:0]  r4_read_data;

    always @(posedge sys_clk or negedge sys_rst_n) begin
      if (sys_rst_n == 1'b0) begin
        r_reg_addr     <=  16'b0;
      end else if (cmd_en == 1'b1) begin
        r_reg_addr     <=  reg_addr;
      end
    end
    
    always @(posedge sys_clk or negedge sys_rst_n) begin
      if (sys_rst_n == 1'b0) begin
        r_divide_clk     <=  9'b0;
        r1_divide_clk    <=  1'b0;
      end else begin
        r_divide_clk     <= r_divide_clk + 1'b1;
        r1_divide_clk    <= r_divide_clk[8];
      end
    end

    assign slow_clk = r1_divide_clk^r_divide_clk[8];     
    
    always @(posedge sys_clk or negedge sys_rst_n) begin
      if (sys_rst_n == 1'b0) begin
        r_write_data_cnt       <=  5'b0;
      end else if (gen_state == IDLE) begin
        r_write_data_cnt       <=  5'b0;
      end else if (slow_clk &&(r_cnt == 3'b11) && (gen_state == I2C_WRITE)) begin
        r_write_data_cnt     <= r_write_data_cnt + 1'b1;
      end
    end 

    always @(posedge sys_clk or negedge sys_rst_n) begin
      if (sys_rst_n == 1'b0) begin
        r_read_data_cnt       <=  6'b0;
      end else if (gen_state == IDLE) begin
        r_read_data_cnt       <=  6'b0;
      end else if (slow_clk && (r_cnt == 3'b11) && (gen_state == I2C_READ)) begin
        r_read_data_cnt     <= r_read_data_cnt + 1'b1;
      end
    end	

    always @( * )
      begin
        next_gen_state = gen_state;
        case (gen_state)
          IDLE : begin
            if (cmd_en)
              next_gen_state = I2C_START;
          end
          I2C_START: begin
            if (slow_clk && (r_cnt == 3'b11)) begin
              next_gen_state = I2C_WRITE;
            end   
          end  
          I2C_WRITE:begin
            if (r_write_data_cnt == 5'd27) begin          
              next_gen_state = I2C_RESTART;
            end 
          end
		  I2C_RESTART:begin
            if (slow_clk && (r_cnt == 3'b11)) begin
              next_gen_state = I2C_READ;
            end		  
		  end
          I2C_READ:begin
            if (r_read_data_cnt == 6'd45) begin          
              next_gen_state = I2C_END;
            end 
          end		  
          I2C_END:begin
            if (slow_clk && (r_cnt == 3'b11)) begin          
              next_gen_state = IDLE;
            end    
          end      
          default : begin
            next_gen_state = IDLE;
          end
        endcase
      end
    
    always @(posedge sys_clk or negedge sys_rst_n) begin
      if (~sys_rst_n) begin
        gen_state <= IDLE;
      end else begin
        gen_state <= next_gen_state;
      end
    end

    always @(posedge sys_clk or negedge sys_rst_n) begin
      if (sys_rst_n == 1'b0) begin
        r_cnt     <=  2'b0;
      end else if (gen_state == IDLE) begin
        r_cnt     <=  2'b0;      
      end else if (slow_clk) begin
        r_cnt     <= r_cnt + 1'b1;
      end
    end 

    always @(posedge sys_clk or negedge sys_rst_n) begin
      if (sys_rst_n == 1'b0) begin
        r_scl     <=  1'b1;      
        r_sda     <=  1'b1;
      end else if (gen_state == I2C_START) begin
        if (slow_clk && (r_cnt == 3'b0)) begin
          r_sda   <= 1'b0;  
        end else if(slow_clk && (r_cnt == 3'b10)) begin
          r_scl   <= 1'b0;
        end 
      end else if (gen_state == I2C_RESTART) begin
        if (slow_clk && (r_cnt == 3'b01)) begin
          r_scl   <= 1'b1; 
        end else if (slow_clk && (r_cnt == 3'b11)) begin
          r_scl   <= 1'b0;		  
        end else if(slow_clk && (r_cnt == 3'b0)) begin
          r_sda   <= 1'b1;
        end else if(slow_clk && (r_cnt == 3'b10)) begin
          r_sda   <= 1'b0;          
        end 		
      end else if (gen_state == I2C_WRITE) begin
        if (slow_clk && (r_cnt == 3'b0)) begin
          case(r_write_data_cnt)
            5'd0 : r_sda <= SLAVE_ADDR[6];
            5'd1 : r_sda <= SLAVE_ADDR[5];
            5'd2 : r_sda <= SLAVE_ADDR[4];
            5'd3 : r_sda <= SLAVE_ADDR[3];
            5'd4 : r_sda <= SLAVE_ADDR[2];
            5'd5 : r_sda <= SLAVE_ADDR[1];
            5'd6 : r_sda <= SLAVE_ADDR[0];
            5'd7 : r_sda <= 1'b0;
            5'd8 : r_sda <= 1'b1;
            5'd9 : r_sda <= r_reg_addr[15];
            5'd10: r_sda <= r_reg_addr[14];
            5'd11: r_sda <= r_reg_addr[13];
            5'd12: r_sda <= r_reg_addr[12];
            5'd13: r_sda <= r_reg_addr[11];
            5'd14: r_sda <= r_reg_addr[10];
            5'd15: r_sda <= r_reg_addr[9];
            5'd16: r_sda <= r_reg_addr[8];
            5'd17: r_sda <= 1'b1;  
            5'd18: r_sda <= r_reg_addr[7];
            5'd19: r_sda <= r_reg_addr[6];
            5'd20: r_sda <= r_reg_addr[5];
            5'd21: r_sda <= r_reg_addr[4];
            5'd22: r_sda <= r_reg_addr[3];
            5'd23: r_sda <= r_reg_addr[2];
            5'd24: r_sda <= r_reg_addr[1];
            5'd25: r_sda <= r_reg_addr[0];
            5'd26: r_sda <= 1'b1;            
            default : r_sda <= 1'b1;        
          endcase        
        end
        if (slow_clk && (r_cnt == 3'b01)) begin
          r_scl   <= 1'b1;
        end else if (slow_clk && (r_cnt == 3'b11)) begin
          r_scl   <= 1'b0;
        end  
      end else if (gen_state == I2C_READ) begin
        if (slow_clk && (r_cnt == 3'b0)) begin
          case(r_read_data_cnt)
            6'd0 : r_sda <= SLAVE_ADDR[6];
            6'd1 : r_sda <= SLAVE_ADDR[5];
            6'd2 : r_sda <= SLAVE_ADDR[4];
            6'd3 : r_sda <= SLAVE_ADDR[3];
            6'd4 : r_sda <= SLAVE_ADDR[2];
            6'd5 : r_sda <= SLAVE_ADDR[1];
            6'd6 : r_sda <= SLAVE_ADDR[0];
            6'd7 : r_sda <= 1'b1;
            6'd8 : r_sda <= 1'b1;
            6'd17: r_sda <= 1'b0;
            6'd26: r_sda <= 1'b0;
            6'd35: r_sda <= 1'b0;			
            default : r_sda <= 1'b1;        
          endcase        
        end
        if (slow_clk && (r_cnt == 3'b01)) begin
          r_scl   <= 1'b1;
        end else if (slow_clk && (r_cnt == 3'b11)) begin
          r_scl   <= 1'b0;
        end		
      end else if (gen_state == I2C_END) begin
        if (slow_clk && (r_cnt == 3'b01)) begin
          r_scl   <= 1'b1;  
        end else if(slow_clk && (r_cnt == 3'b11)) begin
          r_sda   <= 1'b1;
        end else if(slow_clk && (r_cnt == 3'b0)) begin
          r_sda   <= 1'b0;          
        end    
      end
    end
	
    always @(posedge sys_clk or negedge sys_rst_n) begin
      if (sys_rst_n == 1'b0) begin
        r1_read_data     <=  8'b0; 
        r2_read_data     <=  8'b0;
        r3_read_data     <=  8'b0;
        r4_read_data     <=  8'b0;	
      end else if (gen_state == I2C_RESTART) begin	
        r1_read_data     <=  8'b0; 
        r2_read_data     <=  8'b0;
        r3_read_data     <=  8'b0;
        r4_read_data     <=  8'b0;	  
      end else if (gen_state == I2C_READ) begin
        if (slow_clk && (r_cnt == 3'b10)) begin
          case(r_read_data_cnt)
            6'd9  :  r1_read_data <= {r1_read_data[6:0], i2c_sda_in};
            6'd10 :  r1_read_data <= {r1_read_data[6:0], i2c_sda_in};
            6'd11 :  r1_read_data <= {r1_read_data[6:0], i2c_sda_in};
            6'd12 :  r1_read_data <= {r1_read_data[6:0], i2c_sda_in};
            6'd13 :  r1_read_data <= {r1_read_data[6:0], i2c_sda_in};
            6'd14 :  r1_read_data <= {r1_read_data[6:0], i2c_sda_in};
            6'd15 :  r1_read_data <= {r1_read_data[6:0], i2c_sda_in};
			6'd16 :  r1_read_data <= {r1_read_data[6:0], i2c_sda_in};
            6'd18 :  r2_read_data <= {r2_read_data[6:0], i2c_sda_in};
            6'd19 :  r2_read_data <= {r2_read_data[6:0], i2c_sda_in};
            6'd20 :  r2_read_data <= {r2_read_data[6:0], i2c_sda_in};
            6'd21 :  r2_read_data <= {r2_read_data[6:0], i2c_sda_in};
            6'd22 :  r2_read_data <= {r2_read_data[6:0], i2c_sda_in};
            6'd23 :  r2_read_data <= {r2_read_data[6:0], i2c_sda_in};
            6'd24 :  r2_read_data <= {r2_read_data[6:0], i2c_sda_in};
			6'd25 :  r2_read_data <= {r2_read_data[6:0], i2c_sda_in};
            6'd27 :  r3_read_data <= {r3_read_data[6:0], i2c_sda_in};
            6'd28 :  r3_read_data <= {r3_read_data[6:0], i2c_sda_in};
            6'd29 :  r3_read_data <= {r3_read_data[6:0], i2c_sda_in};
            6'd30 :  r3_read_data <= {r3_read_data[6:0], i2c_sda_in};
            6'd31 :  r3_read_data <= {r3_read_data[6:0], i2c_sda_in};
            6'd32 :  r3_read_data <= {r3_read_data[6:0], i2c_sda_in};
            6'd33 :  r3_read_data <= {r3_read_data[6:0], i2c_sda_in};
			6'd34 :  r3_read_data <= {r3_read_data[6:0], i2c_sda_in};
            6'd36 :  r4_read_data <= {r4_read_data[6:0], i2c_sda_in};
            6'd37 :  r4_read_data <= {r4_read_data[6:0], i2c_sda_in};
            6'd38 :  r4_read_data <= {r4_read_data[6:0], i2c_sda_in};
            6'd39 :  r4_read_data <= {r4_read_data[6:0], i2c_sda_in};
            6'd40 :  r4_read_data <= {r4_read_data[6:0], i2c_sda_in};
            6'd41 :  r4_read_data <= {r4_read_data[6:0], i2c_sda_in};
            6'd42 :  r4_read_data <= {r4_read_data[6:0], i2c_sda_in};
            6'd43 :  r4_read_data <= {r4_read_data[6:0], i2c_sda_in};			
          endcase        
        end	   
      end
    end	

    always @(posedge sys_clk or negedge sys_rst_n) begin
      if (sys_rst_n == 1'b0) begin
        r_read_valid     <=  1'b0;    
      end else if ((gen_state == I2C_END) && slow_clk && (r_cnt == 3'b11)) begin
        r_read_valid     <=  1'b1;
      end else if ( slow_clk ) begin
        r_read_valid     <=  1'b0;        
      end
    end    
    
    assign read_valid = r_read_valid & slow_clk;
	assign read_data = {r4_read_data, r3_read_data, r2_read_data, r1_read_data};
    
    assign i2c_sda_out = (r_sda == 1'b0)? 1'b0 : 1'bz;
    assign i2c_scl = r_scl;
        
endmodule              