//******************************************************************************
//
//  File Name:  max_load_monitor.v
//  Version:    1.0
//  Date:       April 28, 2014
//  Model:      max load monitor
//  Function:   loading status monitor and fpga reset control
//  Author:     mechen@altera.com
//  Company:    Altera
//
//*****************************************************************************/

`timescale 1ns / 100ps

module  max_load_monitor
        (
        input   wire  i_reset_n        , // reset
        input   wire  i_system_clk     , // clock
        input   wire  i_slow_clk       , // clock
        input   wire  i_cpu_resetn_gui , // CPU reset via GUI
        input   wire  i_pgm_config     , // FPGA reconfiguration
        input   wire  i_pfl_granted    , // PFL Flash access granting signal
        input   wire  i_fpga_conf_done , // FPGA configuration done
        output  wire  o_max_conf_done  , // MAX loading done
        output  wire  o_max_load       , // MAX loading indication
        output  wire  o_max_load_enable, // MAX loading enable
        output  wire  o_max_error      , // MAX loading error
        output  wire  o_cpu_resetn       // CPU Reset
        );

//--------------------------------------------------
// Signal defination
//--------------------------------------------------

reg               fpga_conf_done_reg1      ; // 
reg               fpga_conf_done_reg2      ; // 
wire              fpga_conf_done_pos       ; // 
reg               slow_clk_reg1            ; // 
reg               slow_clk_reg2            ; // 
wire              slow_clk_neg             ; // 
reg               max_load_enable          ; // 
//reg  [15:0]     max_load_cnt             ; // 0~65535
reg    [18:0]     max_load_cnt             ; // 0~‭524287‬
reg               max_load_error           ; // 
reg    [4:0]      cpu_resetn_cnt           ; // 0~31
reg               cpu_resetn               ; // 

//--------------------------------------------------
// Main function body
//--------------------------------------------------

// Output assignments
assign o_cpu_resetn      = cpu_resetn;
assign o_max_load_enable = max_load_enable;

//---------------------------------------------------------------------------------------
//FPGA/CPU reset
parameter       FILTER_WIDTH = 4            ;
parameter       INIT_BIT     = 0            ;

filter_circuit      #(
                    .FILTER_WIDTH             (FILTER_WIDTH            ), // even number excepts 0, such as 2, 4, 6, 8 ...
                    .INIT_BIT                 (INIT_BIT                )  // initialization bit, 0 or 1
                    )
                    filter_circuit_inst
                    (
                    .i_clk_rst                (!i_reset_n              ), // reset
                    .i_clk                    (i_slow_clk              ), // clock
                    .i_framein                (i_fpga_conf_done        ), // input
                    .o_frameout               (fpga_conf_done          )  // output
                    );

always @(posedge i_system_clk)
begin
fpga_conf_done_reg1 <= fpga_conf_done;
fpga_conf_done_reg2 <= fpga_conf_done_reg1;
end

assign fpga_conf_done_pos = (!fpga_conf_done_reg2) & fpga_conf_done_reg1;       // posedge of i_fpga_conf_done

always @(posedge i_system_clk)
begin
  if(!i_cpu_resetn_gui)
     cpu_resetn_cnt <= 5'd0;
  else if(fpga_conf_done_pos)
     cpu_resetn_cnt <= 5'd0;
  else if(cpu_resetn_cnt == 5'd31)
     cpu_resetn_cnt <= cpu_resetn_cnt;
  else
     cpu_resetn_cnt <= cpu_resetn_cnt + 1'd1;
end

always @(posedge i_system_clk)
begin
  if(cpu_resetn_cnt == 5'd0)
     cpu_resetn <= 1'b0;
  else if(cpu_resetn_cnt == 5'd31)
     cpu_resetn <= 1'bZ;
  else
     cpu_resetn <= cpu_resetn;
end

//---------------------------------------------------------------------------------------
//LED control and unused output signals assignment

always @(posedge i_system_clk or negedge i_reset_n)
begin
  if(!i_reset_n)
     max_load_enable <= 1'b1;
  else if(!i_pgm_config)                                                        // MAX V load enable trigger
     max_load_enable <= 1'b1;
  else if(fpga_conf_done_pos)                                                   // de-assert max5 load enable once fpga configuration is done, to avoid led missing flashing for jtag download way
     max_load_enable <= 1'b0;
  //else if(max_load_cnt == 16'd65535)                                          // de-assert max5 load enable if it cannot successfully configure fpga after certain duration
  else if(max_load_cnt == 19'd524287)
     max_load_enable <= 1'b0;
  else
     max_load_enable <= max_load_enable;
end

always @(posedge i_system_clk or negedge i_reset_n)
begin
  if(!i_reset_n)
     begin
     slow_clk_reg1 <= 1'b0;
     slow_clk_reg2 <= 1'b0;
     end
  else
     begin
     slow_clk_reg1 <= i_slow_clk;
     slow_clk_reg2 <= slow_clk_reg1;
     end
end

assign slow_clk_neg = (!slow_clk_reg1) & slow_clk_reg2;                         // negedge

//always @(posedge i_system_clk or negedge i_reset_n)
//begin
//  if(!i_reset_n)
//     max_load_cnt <= 16'd0;
//  else if(!i_pgm_config)                                                      // clear counter at the beginning of each loading round
//     max_load_cnt <= 16'd0;
//  else if(fpga_conf_done_pos)
//     max_load_cnt <= 16'd0;
//  else if(max_load_enable & slow_clk_neg)
//     max_load_cnt <= max_load_cnt + 1'd1;
//  else
//     max_load_cnt <= max_load_cnt;
//end

always @(posedge i_system_clk or negedge i_reset_n)
begin
  if(!i_reset_n)
     max_load_cnt <= 19'd0;
  else if(!i_pgm_config)                                                        // clear counter at the beginning of each loading round
     max_load_cnt <= 19'd0;
  else if(fpga_conf_done_pos)
     max_load_cnt <= 19'd0;
  else if(max_load_enable & slow_clk_neg)
     max_load_cnt <= max_load_cnt + 1'd1;
  else
     max_load_cnt <= max_load_cnt;
end

always @(posedge i_system_clk or negedge i_reset_n)
begin
  if(!i_reset_n)
     max_load_error <= 1'b0;
  else if(!i_pgm_config)
     max_load_error <= 1'b0;
  else if(fpga_conf_done_reg2)                                                  // MAX V won't reconfigure FPGA right after itself reconfiguration via programmer
     max_load_error <= 1'b0;                                                    // When FPGA configuration done status is active(high)
  //else if(max_load_cnt == 16'd65535)                                          // Assert max_load_error once counter overflow, timeout
  else if(max_load_cnt == 19'd524287)                                           // Assert max_load_error once counter overflow, timeout
     max_load_error <= 1'b1;
  else
     max_load_error <= max_load_error;
end

assign o_max_conf_done = !fpga_conf_done ;
assign o_max_load      = (!max_load_enable) || (!i_pfl_granted);
assign o_max_error     = !max_load_error;

endmodule
//------------------------------------------------------------------------------
// end of file
//------------------------------------------------------------------------------