//******************************************************************************
//
//  File Name:  max_pfl_ctrl.v
//  Version:    1.0
//  Date:       April 28, 2014
//  Model:      PFL controller
//  Function:   Manage PFL controller and monitor max loading status as well as FPGA reset control
//  Author:     mechen@altera.com
//  Company:    Altera
//
//*****************************************************************************/

`timescale 1ns / 100ps

module  max_pfl_ctrl
        (
        //Reset and Clock
        input   wire        i_reset_n           , // reset
        input   wire        i_system_clk        , // clock_factory_load
        input   wire        i_slow_clk          , // clock

        //FPGA configuration control and status
        input   wire        i_cpu_resetn_gui    , // CPU reset via GUI
        input   wire        i_fpga_reconfign_gui, // FPGA reconfiguration via GUI
        input   wire [3:0]  i_page_sel_gui      , // Page selection via GUI
        input   wire        i_factory_load      , // Factory load selection
        input   wire        i_pgm_sel           , // Flash page selection button
        input   wire        i_pgm_config        , // FPGA reconfiguration
        input   wire        i_fpga_conf_done    , // FPGA configuration done
        input   wire        i_fpga_nstatus      , // FPGA configuration status

        //Flash interface
        output  wire [25:0] o_fm_a              , // Parallel Flash interface, also connect to FPGA
        inout   wire [15:0] io_fm_d             , // Parallel Flash interface, also connect to FPGA
        output  wire [1:0]  o_flash_cen         , // Parallel Flash interface, also connect to FPGA
        input   wire [1:0]  i_flash_rdybsyn     , // Parallel Flash interface, also connect to FPGA, not used currently
        output  wire        o_flash_advn        , // Parallel Flash interface, also connect to FPGA
        output  wire        o_flash_clk         , // Parallel Flash interface, also connect to FPGA
        output  wire        o_flash_oen         , // Parallel Flash interface, also connect to FPGA
        output  wire        o_flash_resetn      , // Parallel Flash interface, also connect to FPGA
        output  wire        o_flash_wen         , // Parallel Flash interface, also connect to FPGA

        //FPP interface	
        output  wire        o_avst_valid        , // AVST configuration valid
		input   wire        i_avst_ready        , // AVST configuration ready
		output  wire [15:0] o_avst_data         , // AVST configuration data bus
        output  wire        o_avst_clk          , // AVST configuration clock
        output  wire        o_fpga_nconfig      , // AVST configuration config control, low active

        //MAX V loading monitor status
        output  wire        o_max_conf_done     , // MAX loading done
        output  wire        o_max_load          , // MAX loading indication
        output  wire        o_max_error         , // MAX loading error
        output  wire [2:0]  o_fpga_pgm          , // FPGA programming page selection
        output  wire [2:0]  o_pgm_led           , // Page selection LED indication
		output  wire        o_pgm_sel_pb_trig	, // Page_sel pushbutton
        output  wire        o_cpu_resetn          // CPU Reset
        );

//--------------------------------------------------
// Signal defination
//--------------------------------------------------

parameter       FILTER_WIDTH = 4            ;
parameter       INIT_BIT     = 1            ;

wire  [2:0]     fpga_pgm                    ;
wire  [2:0]     pfl_pgm                     ;
wire            pgm_config                  ;
wire            pgm_config_ctrl             ;
wire            flash_cen                   ;
wire            pfl_flash_access_granted    ;
wire            max_load_enable             ;
   
wire  [15:0]    fpga_config_d_flash         ;
wire            fpga_dclk_flash             ;
wire            fpga_nconfig_flash          ;
reg             avst_ready_reg1             ;
reg             avst_ready_reg2             ;
wire            avst_clk_temp               ; 

//--------------------------------------------------
// Main function body
//--------------------------------------------------

assign o_fpga_pgm = fpga_pgm;

//---------------------------------------------------------------------------------------
//Programming image selection
//
max_pgm_sel         max_pgm_sel_inst
                    (
                    .i_reset_n                (i_reset_n               ), // reset
                    .i_slow_clk               (i_slow_clk              ), // clock
                    .i_pgm_sel_disable        (max_load_enable         ), // Disable page changed during configuration
                    .i_page_sel_gui           (i_page_sel_gui          ), // Page selection via GUI
                    .i_factory_load           (i_factory_load          ), // Factory load setting
                    .i_pgm_sel                (i_pgm_sel               ), // Programming page select
                    .o_pgm_led                (o_pgm_led               ), // LED indication
                    .o_pgm_sel_pb_trig        (o_pgm_sel_pb_trig	   ), // Page_sel pushbutton
                    .o_fpga_pgm               (fpga_pgm                )  // FPGA programming page selection
                    );

//---------------------------------------------------------------------------------------
//Programming signal filter control
//
filter_circuit      #(
                    .FILTER_WIDTH             (FILTER_WIDTH            ), // even number excepts 0, such as 2, 4, 6, 8 ...
                    .INIT_BIT                 (INIT_BIT                )  // initialization bit, 0 or 1
                    )
                    filter_circuit_inst
                    (
                    .i_clk_rst                (!i_reset_n              ), // reset
                    .i_clk                    (i_slow_clk              ), // clock
                    .i_framein                (i_pgm_config            ), // input
                    .o_frameout               (pgm_config              )  // output
                    );

assign pgm_config_ctrl = pgm_config & i_fpga_reconfign_gui;

//--------------
//PFL interface
//--------------				
assign o_avst_clk = avst_clk_temp;					
					
pfl2                pfl_x32_inst
                    (
                    .avst_valid               (o_avst_valid            ),
                    .avst_ready               (i_avst_ready            ),
                    .avst_data                (o_avst_data             ),
                    .avst_clk                 (avst_clk_temp           ),
                    .flash_addr               (o_fm_a                  ),
                    //.flash_clk              (o_flash_clk             ),
                    .flash_data               (io_fm_d                 ),
                    //.flash_nadv             (o_flash_advn            ),
                    .flash_nce                (flash_cen               ),
                    .flash_noe                (o_flash_oen             ),
                    .flash_nreset             (o_flash_resetn          ),
                    .flash_nwe                (o_flash_wen             ),
                    .fpga_conf_done           (i_fpga_conf_done        ),
                    .fpga_nconfig             (fpga_nconfig_flash      ),
                    .fpga_nstatus             (i_fpga_nstatus          ),
                    .fpga_pgm                 (pfl_pgm                 ),
                    .pfl_clk                  (i_system_clk            ),
                    .pfl_flash_access_granted (pfl_flash_access_granted),
                    .pfl_flash_access_request (pfl_flash_access_granted),
                    .pfl_nreconfigure         (pgm_config_ctrl         ),
                    .pfl_nreset               (i_reset_n               )
                    );					

assign o_flash_cen = (fpga_pgm == 3'b010) ? {flash_cen, 1'bz} : {1'bz, flash_cen};
assign pfl_pgm = (fpga_pgm == 3'b010) ? 3'b000 : fpga_pgm;
assign o_flash_clk = 1'b0;
assign o_flash_advn = 1'b0;

//---------------------------------------------------------------------------------------
//FPGA configuration interface
//
//assign o_fpga_config_d = max_load_enable ? fpga_config_d_flash : 32'hZZZZZZZZ;
assign o_fpga_config_d = max_load_enable ? fpga_config_d_flash : 16'hZZZZ;
assign o_fpga_dclk     = max_load_enable ? fpga_dclk_flash     : 1'b0;
assign o_fpga_nconfig  = max_load_enable ? fpga_nconfig_flash  : 1'b1;

//---------------------------------------------------------------------------------------
//MAX loading status monitor and fpga reset control
//
max_load_monitor    max_load_monitor_inst
                    (
                    .i_reset_n                (i_reset_n               ), // reset
                    .i_system_clk             (i_system_clk            ), // clock
                    .i_slow_clk               (i_slow_clk              ), // clock
                    .i_cpu_resetn_gui         (i_cpu_resetn_gui        ), // CPU reset via GUI
                    .i_pgm_config             (pgm_config_ctrl         ), // FPGA reconfiguration
                    .i_pfl_granted            (pfl_flash_access_granted), // PFL Flash access granting signal
                    .i_fpga_conf_done         (i_fpga_conf_done        ), // FPGA configuration done
                    .o_max_conf_done          (o_max_conf_done         ), // MAX loading done
                    .o_max_load               (o_max_load              ), // MAX loading indication
                    .o_max_load_enable        (max_load_enable         ), // MAX loading enable
                    .o_max_error              (o_max_error             ), // MAX loading error
                    .o_cpu_resetn             (o_cpu_resetn            )  // CPU Reset, low active
                    );

endmodule
//------------------------------------------------------------------------------
// end of file
//------------------------------------------------------------------------------