library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity max5 is
  port 
  (
    ---------- Other PBs & LEDs ----------
    cpu_resetn          : inout std_logic;                     -- Reset PB Max V & Stratix V
    max_resetn          : in    std_logic;                     -- Max V Reset PB

    -----------clocks-----------
    clk_50m_max5        : in    std_logic;                     -- 50MHz
    clk_config          : in    std_logic;                     -- 100MHz, configuration clock
    max_osc_clk_1       : in    std_logic;                     -- 125MHz, AVST configuration reference clock

    clk0_oen            : out   std_logic;                     -- clock oen
    clk0_rstn           : out   std_logic;                     -- clock rstn
    clk0_finc           : out   std_logic;                     -- clock frequency inc
    clk0_fdec           : out   std_logic;                     -- clock frequency dec
    si516_fs            : out   std_logic;                     -- Si516 FS
  
  -----------I2C--------- 
    i2c_scl             : inout std_logic;                     -- Programmable clocks I2C serial clock
    i2c_sda             : inout std_logic;                     -- Programmable clocks I2C data bus 
  
    -----------fpga----------- (others => 'Z')
    fpga_avst_clk       : out   std_logic;                     -- AVST clock
    fpga_avst_ready     : in    std_logic;                     -- SDM_IO8
    fpga_avst_valid     : out   std_logic;                     -- SDM_IO11
    fpga_config_d       : out   std_logic_vector(15 downto 0); -- FPGA parallel configuration interface
    fpga_nconfig        : out   std_logic;                     -- FPGA parallel configuration interface
    fpga_nstatus        : in    std_logic;                     -- FPGA parallel configuration interface
    fpga_conf_done      : in    std_logic;                     -- FPGA parallel configuration interface
    fpga_cvp_done       : in    std_logic;                     -- FPGA parallel configuration interface
    fpga_pr_done        : in    std_logic;                     -- Not used currently
    fpga_pr_error       : in    std_logic;                     -- Not used currently
    fpga_pr_request     : out   std_logic := 'Z';              -- Not used currently
    fpga_init_done      : in    std_logic;                     -- SDM_IO0
    
    ---------- flash ----------
    fm_a                : out   std_logic_vector(26 downto 1); -- Parallel Flash interface, also connect to FPGA 
    fm_d                : inout std_logic_vector(15 downto 0); -- Parallel Flash interface, also connect to FPGA 
    flash_cen           : out   std_logic_vector( 1 downto 0); -- Parallel Flash interface, also connect to FPGA
    flash_rdybsyn       : in    std_logic_vector( 1 downto 0); -- Parallel Flash interface, also connect to FPGA, not used currently
    flash_advn          : out   std_logic;                     -- Parallel Flash interface, also connect to FPGA
    flash_clk           : out   std_logic;                     -- Parallel Flash interface, also connect to FPGA
    flash_oen           : out   std_logic;                     -- Parallel Flash interface, also connect to FPGA
    flash_resetn        : out   std_logic;                     -- Parallel Flash interface, also connect to FPGA
    flash_wen           : out   std_logic;                     -- Parallel Flash interface, also connect to FPGA
    
    ----------- Configuration -----------
    factory_load        : in    std_logic;                     -- DIPSWITCH - Load Factory or User Design at Power-up
    pgm_config          : in    std_logic;                     -- PB to Configure FPGA via FPP
    pgm_sel             : in    std_logic;                     -- PB to toggle PGM_LEDs
    pgm_led             : out   std_logic_vector(2 downto 0);  -- LED configuration will indicate flash page for FPP
    max_conf_done       : out   std_logic;                     -- LED - ON after PFL Configuration PASS
    max_error           : out   std_logic;                     -- LED - ON if PFL Configuration FAIL
    max_load            : out   std_logic;                     -- LED - ON while loading FPP design from flash
    msel                : in    std_logic_vector(2 downto 0);  -- DIPSWITCH - Stratix V MSEL pins
                                                               -- MSEL[0] -> SDM_IO5
                                                               -- MSEL[1] -> SDM_IO7
                                                               -- MSEL[2] -> SDM_IO9
    
    ---------- max5-S10 bus ---------
    max5_ben            : in    std_logic_vector(3 downto 0);  -- FPGA to MAX V interface for registers access
    max5_clk            : out   std_logic;                     -- FPGA to MAX V interface for registers access
    max5_csn            : out   std_logic;                     -- FPGA to MAX V interface for registers access
    max5_oen            : in    std_logic;                     -- FPGA to MAX V interface for registers access
    max5_wen            : in    std_logic;                     -- FPGA to MAX V interface for registers access
    
    ---------- On-Board USB-Blaster ----------
    usb_max5_clk        : in    std_logic;                     -- Not used currently
    usb_cfg             : inout std_logic_vector(14 downto 0); -- Not used currently
    
    ---------- Temp Sense ----------
    overtempn           : in    std_logic;                     -- Over temperature, MAX1619 
    overtemp            : out   std_logic;                     -- Fan control over temperature
    tsense_alertn       : in    std_logic;                     -- alert, MAX1619

    ---------- FMC daughter card presentation ----------
    fmca_prsntn         : in    std_logic;                     -- FMCA presented or not
  
  ---------- SDI control ----------
    sdi_mf2_mute        : out   std_logic;                     -- SDI MF2
    sdi_mf1_auto_sleep  : out   std_logic;                     -- SDI MF1
    sdi_mf0_bypass      : out   std_logic;                     -- SDI MF0
    sdi_tx_sd_hdn       : out   std_logic                      -- SDI SD or HD mode
    );
end entity;

architecture rtl of max5 is

constant MAX_VER : std_logic_vector(15 downto 0) := x"000A";
constant PROJ_ID : std_logic_vector(15 downto 0) := x"AD75";

signal   reset_n                  : std_logic;

signal   clk_pls_n                : std_logic;
signal   clk_pls_p                : std_logic;
signal   clk_pls_p_deley          : std_logic;
signal   slow_clk                 : std_logic;
signal   mv_clk_12m5              : std_logic;
signal   mv_clk_25m               : std_logic;

signal   fan_cont_out             : std_logic;

signal   diff_raw_int             : std_logic_vector(23 downto 0) := x"000000";
signal   single_raw_int           : std_logic_vector(23 downto 0) := x"000000";

signal   fpga_pgm_int             : std_logic_vector(2 downto 0);
signal   cpu_resetn_gui           : std_logic;
signal   fpga_reconfign_gui       : std_logic;
signal   page_sel_gui             : std_logic_vector(3 downto 0);
signal   control_gui              : std_logic_vector(7 downto 0);
signal   temp_data_gui            : std_logic_vector(7 downto 0);
signal   pgm_sel_pb_trig          : std_logic;
signal   i2c_inf_sw_gui           : std_logic;
signal   channel_no               : std_logic_vector(2 downto 0);
signal   channel_data             : std_logic_vector(31 downto 0);
signal   over_tempn1              : std_logic;
signal   over_tempn2              : std_logic;
signal   power_off_en             : std_logic;

component   reset_generator is
            port 
            (
            clk               : in  std_logic;
            reset_n_in        : in  std_logic;
            reset_n_out       : out std_logic
            );
end component;

component   slow_clocks is
            port
            (
            i_clk               : in  std_logic;
            i_reset_n           : in  std_logic;
            o_clk_pls_n         : out std_logic;
            o_clk_pls_p         : out std_logic;
            o_clk_pls_p_deley   : out std_logic;
            o_slow_clk          : out std_logic;
            o_clk_div2          : out std_logic;
            o_clk_div4          : out std_logic
            );
end component;

component   max_pfl_ctrl is
           port
           (
           --Reset and Clock
           i_reset_n           : in    std_logic;                     -- reset
           i_system_clk        : in    std_logic;                     -- clocki_factory_load
           i_slow_clk          : in    std_logic;                     -- clock
           --FPGA configuration control and status
           i_cpu_resetn_gui    : in    std_logic;                     -- CPU reset via GUI
           i_fpga_reconfign_gui: in    std_logic;                     -- FPGA reconfiguration via GUI
           i_page_sel_gui      : in    std_logic_vector(3 downto 0);  -- Page selection via GUI
           i_factory_load      : in    std_logic;                     -- Factory load selection
           i_pgm_sel           : in    std_logic;                     -- Flash page selection button
           i_pgm_config        : in    std_logic;                     -- FPGA reconfiguration
           i_fpga_conf_done    : in    std_logic;                     -- FPGA configuration done
           i_fpga_nstatus      : in    std_logic;                     -- FPGA configuration status
           --Flash interface
           o_fm_a              : out   std_logic_vector(25 downto 0); -- Parallel Flash interface, also connect to FPGA
           io_fm_d             : inout std_logic_vector(15 downto 0); -- Parallel Flash interface, also connect to FPGA
           o_flash_cen         : out   std_logic_vector( 1 downto 0); -- Parallel Flash interface, also connect to FPGA
           i_flash_rdybsyn     : in    std_logic_vector( 1 downto 0); -- Parallel Flash interface, also connect to FPGA, not used currently
           o_flash_advn        : out   std_logic;                     -- Parallel Flash interface, also connect to FPGA
           o_flash_clk         : out   std_logic;                     -- Parallel Flash interface, also connect to FPGA
           o_flash_oen         : out   std_logic;                     -- Parallel Flash interface, also connect to FPGA
           o_flash_resetn      : out   std_logic;                     -- Parallel Flash interface, also connect to FPGA
           o_flash_wen         : out   std_logic;                     -- Parallel Flash interface, also connect to FPGA
           --FPP interface
           o_avst_data         : out   std_logic_vector(15 downto 0); -- FPP configuration data bus
           o_avst_valid        : out   std_logic;                     -- valid
           i_avst_ready        : in    std_logic;                     -- fpga_avst_ready     
           o_avst_clk          : out   std_logic;                     -- FPP configuration clock
           o_fpga_nconfig      : out   std_logic;                     -- FPP configuration config control, low active
           --MAX V loading monitor status
           o_max_conf_done     : out   std_logic;                     -- MAX loading done
           o_max_load          : out   std_logic;                     -- MAX loading indication
           o_max_error         : out   std_logic;                     -- MAX loading error
           o_fpga_pgm          : out   std_logic_vector(2  downto 0); -- FPGA programming page selection
           o_pgm_led           : out   std_logic_vector(2  downto 0); -- Page selection LED indication
           o_pgm_sel_pb_trig   : out   std_logic;                     -- Page_sel pushbutton
           o_cpu_resetn        : out   std_logic                      -- CPU Reset
           );
end component;

component   fan_cont is
            port
            (
            reset_n           : in std_logic;
            clk               : in std_logic;
            clk_pls_p         : in std_logic;
            data_out_t_r      : in std_logic_vector(7 downto 0);
            fan_cont_out      : out std_logic
            );
end component;  

component   q_sys is
            port
            (
            clk_clk                               : in    std_logic := 'X'; -- clk
            reset_reset_n                         : in    std_logic := 'X'; -- reset_n
            system_max_id_0_version_export        : in    std_logic_vector(15 downto 0);
            system_max_id_0_project_id_export     : in    std_logic_vector(15 downto 0);
            system_max_id_0_diff_raw_export       : in    std_logic_vector(23 downto 0);
            system_max_id_0_single_raw_export     : in    std_logic_vector(23 downto 0);
            system_max_id_0_control_gui_export    : out   std_logic_vector(7  downto 0);
            system_max_id_0_page_sel_pb_export    : in    std_logic;
            system_max_id_0_fpga_pgm_export       : in    std_logic_vector(2  downto 0);
            system_max_id_0_cpu_resetn_export     : out   std_logic;
            system_max_id_0_fpga_reconfign_export : out   std_logic;
            system_max_id_0_page_sel_export       : out   std_logic_vector(3  downto 0);
            system_max_id_0_temp_data_export      : out   std_logic_vector(7  downto 0);
            system_max_id_0_i2c_inf_sw_export     : out   std_logic;
            pio_ch_data_external_connection_export: in    std_logic_vector(31 downto 0);
            pio_ch_no_external_connection_export  : out   std_logic_vector(2 downto 0);
            opencores_i2c_0_slow_clock_slow_clock : in    std_logic;
            opencores_i2c_0_over_temp_export      : in    std_logic;      
            opencores_i2c_0_export_0_scl_pad_io   : inout std_logic;
            opencores_i2c_0_export_0_sda_pad_io   : inout std_logic
            --opencores_i2c_0_export_0_i2c_sel     : in    std_logic
            );
end component q_sys;

component tsd_read
    port
  (
        sys_clk               :in      std_logic;
        sys_rst_n             :in      std_logic;
        i2c_sda_in            :in      std_logic;
        i2c_sda_out           :out     std_logic;
        i2c_scl               :out     std_logic;
        over_tempn1           :out     std_logic;
        over_tempn2           :out     std_logic;
        channel_no            :in      std_logic_vector( 2 downto 0);
        channel_data          :out     std_logic_vector(31 downto 0)
  );
end component;

-----------------------------------------------------------------------------------------
--Main Function

begin

reset_generator_inst :  reset_generator port map
                        (
                        clk         => clk_50m_max5,
                        reset_n_in  => max_resetn,
                        reset_n_out => reset_n
                        );

slow_clocks_inst :      slow_clocks port map
                        (
                        i_clk             => clk_50m_max5,
                        i_reset_n         => reset_n,
                        o_clk_pls_n       => clk_pls_n,
                        o_clk_pls_p       => clk_pls_p,
                        o_clk_pls_p_deley => clk_pls_p_deley,
                        o_slow_clk        => slow_clk,
                        o_clk_div2        => mv_clk_25m,
                        o_clk_div4        => mv_clk_12m5
                        );

-----------------------------------------------------------------------------------------
--MAX V PFL Contorl

max_pfl_ctrl_inst:      max_pfl_ctrl port map
                       (
                       --Reset and Clock
                       i_reset_n            => reset_n           , -- reset
                       i_system_clk         => clk_50m_max5      , -- clocki_factory_load, mv_clk_50
                       i_slow_clk           => slow_clk          , -- clock
                       --FPGA configuration control and status
                       i_cpu_resetn_gui     => cpu_resetn_gui    , -- CPU reset via GUI
                       i_fpga_reconfign_gui => fpga_reconfign_gui, -- FPGA reconfiguration via GUI
                       i_page_sel_gui       => page_sel_gui      , -- Page selection via GUI
                       i_factory_load       => factory_load      , -- Factory load selection
                       i_pgm_sel            => pgm_sel           , -- Flash page selection button
                       i_pgm_config         => pgm_config        , -- FPGA reconfiguration
                       i_fpga_conf_done     => fpga_conf_done    , -- FPGA configuration done
                       i_fpga_nstatus       => fpga_nstatus      , -- FPGA configuration status
                       --Flash interface
                       o_fm_a               => fm_a(26 downto 1) , -- Parallel Flash interface, also connect to FPGA
                       io_fm_d              => fm_d              , -- Parallel Flash interface, also connect to FPGA
                       o_flash_cen          => flash_cen         , -- Parallel Flash interface, also connect to FPGA
                       i_flash_rdybsyn      => flash_rdybsyn     , -- Parallel Flash interface, also connect to FPGA, not used currently
                       o_flash_advn         => flash_advn        , -- Parallel Flash interface, also connect to FPGA
                       o_flash_clk          => flash_clk         , -- Parallel Flash interface, also connect to FPGA
                       o_flash_oen          => flash_oen         , -- Parallel Flash interface, also connect to FPGA
                       o_flash_resetn       => flash_resetn      , -- Parallel Flash interface, also connect to FPGA
                       o_flash_wen          => flash_wen         , -- Parallel Flash interface, also connect to FPGA
                       --FPP interface
                       o_avst_data          => fpga_config_d     , -- data
                       o_avst_valid         => fpga_avst_valid   , -- valid
                       i_avst_ready         => fpga_avst_ready   , -- fpga_avst_ready
                       o_avst_clk           => fpga_avst_clk     , -- fpga_avst_clk        
                       o_fpga_nconfig       => fpga_nconfig      , -- FPP configuration config control, low active
                       --MAX V loading monitor status
                       o_max_conf_done      => max_conf_done     , -- MAX loading done
                       o_max_load           => max_load          , -- MAX loading indication
                       o_max_error          => max_error         , -- MAX loading error
                       o_fpga_pgm           => fpga_pgm_int      , -- FPGA programming page selection
                       o_pgm_led            => pgm_led           , -- Page selection LED indication
                       o_pgm_sel_pb_trig    => pgm_sel_pb_trig   , -- Page_sel pushbutton
                       o_cpu_resetn         => cpu_resetn          -- CPU Reset
                       );

-- -----------------------------------------------------------------------------------------
-- --Temperature
fan_cont_inst:          fan_cont port map
                        (
                        reset_n       => reset_n,
                        clk           => clk_50m_max5,
                        clk_pls_p     => clk_pls_p,
                        data_out_t_r  => temp_data_gui,         -- unsigned integer from 0 to 127, default 48
                        fan_cont_out  => fan_cont_out
                        );

overtemp <= '1' when ((tsense_alertn = '0') or (over_tempn1 = '0')) else (not fan_cont_out); -- Fan control

-----------------------------------------------------------------------------------------
--Qsys block for i2c interface and build info access

q_sys_u0:   q_sys port map
            (
            clk_clk                                => clk_50m_max5      , -- //                            clk.clk
            reset_reset_n                          => reset_n           , -- //                          reset.reset_n
            system_max_id_0_version_export         => MAX_VER           , -- //        system_max_id_0_version.export
            system_max_id_0_project_id_export      => PROJ_ID           , -- //     system_max_id_0_project_id.export
            system_max_id_0_diff_raw_export        => diff_raw_int      , -- //       system_max_id_0_diff_raw.export
            system_max_id_0_single_raw_export      => single_raw_int    , -- //     system_max_id_0_single_raw.export
            system_max_id_0_page_sel_pb_export     => pgm_sel_pb_trig  , -- //     system_max_id_0_page_sel_pb_export
            system_max_id_0_fpga_pgm_export        => fpga_pgm_int      , -- //       system_max_id_0_fpga_pgm.export
            system_max_id_0_cpu_resetn_export      => cpu_resetn_gui    , -- //     system_max_id_0_cpu_resetn.export
            system_max_id_0_fpga_reconfign_export  => fpga_reconfign_gui, -- // system_max_id_0_fpga_reconfign.export
            system_max_id_0_page_sel_export        => page_sel_gui      , -- //       system_max_id_0_page_sel.export
            system_max_id_0_control_gui_export     => control_gui       , -- //    system_max_id_0_control_gui.export
            system_max_id_0_temp_data_export       => temp_data_gui     , -- //      system_max_id_0_temp_data.export
            system_max_id_0_i2c_inf_sw_export      => i2c_inf_sw_gui    , -- //     system_max_id_0_i2c_inf_sw.export
            pio_ch_data_external_connection_export => channel_data      ,
            pio_ch_no_external_connection_export   => channel_no        ,
            opencores_i2c_0_slow_clock_slow_clock  => slow_clk          , -- // opencores_i2c_0_slow_clock_slow_clock
            opencores_i2c_0_over_temp_export       => power_off_en      , -- //       opencores_i2c_0_over_temp_export
            opencores_i2c_0_export_0_scl_pad_io    => i2c_scl           , -- //       opencores_i2c_0_export_0.scl_pad_io
            opencores_i2c_0_export_0_sda_pad_io    => i2c_sda             -- //                               .sda_pad_io
            --opencores_i2c_0_export_0_i2c_sel     => i2c_inf_sw_gui      -- //                               .i2c_sel
            );

power_off_en <= overtempn and over_tempn2;

tsd_read_i : tsd_read port map 
(
    sys_clk               =>  clk_50m_max5          ,  
    sys_rst_n             =>  reset_n               ,
    i2c_sda_in            =>  max5_oen              ,
    i2c_sda_out           =>  max5_csn              ,
    i2c_scl               =>  max5_clk              ,
    over_tempn1           =>  over_tempn1           ,          --Set Fan speed to 100%
    over_tempn2           =>  over_tempn2           ,          --Sequence off power
    channel_no            =>  channel_no            ,
    channel_data          =>  channel_data          
);      
-----------------------------------------------------------------------------------------
--Output signals assignment
fpga_pr_request    <= 'Z'; 
usb_cfg            <= (OTHERS => 'Z');
sdi_mf2_mute       <= 'Z';
sdi_mf1_auto_sleep <= 'Z';
sdi_mf0_bypass     <= 'Z';
sdi_tx_sd_hdn      <= 'Z';
clk0_oen           <= '0';
clk0_rstn          <= 'Z';
clk0_finc          <= 'Z';
clk0_fdec          <= 'Z';
si516_fs           <= 'Z';

end rtl;
